import aiohttp
from pprint import pprint

from .get_data import get_data
from ..DomoAuth import DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def get_grants(full_auth: DomoFullAuth,
                     debug: bool = False, log_results: bool = False,
                     session: aiohttp.ClientSession = None):
    try:
        is_close_session = False

        if not session:
            session = aiohttp.ClientSession()
            is_close_session = True

        url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/authorities'

        if debug:
            print(url)

        return await get_data(auth=full_auth,
                              url=url,
                              method='GET',
                              log_results=log_results,
                              debug=debug,
                              session=session)

    finally:
        if is_close_session:
            await session.close()



import asyncio
import datetime as dt
import io
import json
from dataclasses import dataclass, field
from enum import Enum, auto

import aiohttp
import pandas as pd

from ..utils.chunk_execution import chunk_list
from ..utils.DictDot import DictDot
from . import DomoCertification as dmdc
from .DomoAuth import DomoDeveloperAuth, DomoFullAuth
from .routes import pdp_routes


@dataclass
class PDP_Policy():
    dataset_id: str
    filter_group_id: str
    name: str
    resources: list
    parameters: list

    @classmethod
    def _from_json(cls, json_obj):
        dd = DictDot(json_obj)

        return cls(dataset_id=dd.dataSourceId,
                   filter_group_id=dd.filterGroupId,
                   name=dd.name,
                   resources=dd.resources,
                   parameters=dd.parameters)

    @staticmethod
    def generate_parameter_simple(column_name, column_values_list, operator='EQUALS', ignore_case: bool = True):
        return pdp_routes.generate_policy_parameter_simple(column_name=column_name,
                                                           column_values_list=column_values_list,
                                                           operator=operator,
                                                           ignore_case=ignore_case)

    @staticmethod
    def generate_body(policy_name, dataset_id, parameters_list, policy_id=None, user_ids=None, group_ids=None, virtual_user_ids=None):
        return pdp_routes.generate_policy_body(policy_name=policy_name,
                                               dataset_id=dataset_id,
                                               parameters_list=parameters_list,
                                               policy_id=policy_id,
                                               user_ids=user_ids,
                                               group_ids=group_ids,
                                               virtual_user_ids=virtual_user_ids)

    @classmethod
    async def update_policy(cls, full_auth: DomoFullAuth,
                            dataset_id: str,
                            policy_definition: dict, # body sent to the API (uses camelCase instead of snake_case)
                            debug: bool = False):
        
        print(policy_definition)
        
        if policy_definition.get('filterGroupId'):
            
            res = await pdp_routes.update_policy(full_auth=full_auth,
                                                 dataset_id=dataset_id,
                                                 filter_group_id=policy_definition.get(
                                                     'filterGroupId'),
                                                 body=policy_definition,
                                                 debug=debug)
            return cls._from_json(res.response)
        else:
            res = await pdp_routes.create_policy(full_auth=full_auth,
                                                 dataset_id=dataset_id,
                                                 body=policy_definition,
                                                 debug=debug)

            return cls._from_json(res.response)
    
class Dataset_PDP_Policies:

    def __init__(self, dataset):
        self.dataset = dataset
        self.policies: list = []

    async def get_policies(self, full_auth: DomoFullAuth = None, dataset_id: str = None, debug: bool = False):

        dataset_id = dataset_id or self.dataset.id
        full_auth = full_auth or self.dataset.full_auth

        res = await pdp_routes.get_pdp_policies(full_auth=full_auth, dataset_id=dataset_id, debug=debug)

        if debug:
            print("debug")
            print(res.status)
            print(res.response)

        if res.status == 200:
            domo_policy = [PDP_Policy._from_json(
                policy_obj) for policy_obj in res.response]
            self.policies = domo_policy
            return domo_policy



import aiohttp
import pandas as pd
import io
from pprint import pprint

from .get_data import get_data, looper
from ..DomoAuth import DomoDeveloperAuth, DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def get_pdp_policies(full_auth: DomoFullAuth, dataset_id: str, debug: bool = False) -> ResponseGetData:
    url = f'http://{full_auth.domo_instance}.domo.com/api/query/v1/data-control/{dataset_id}/filter-groups/'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        debug=debug,
    )
    return res


####  PDP ####
def generate_policy_parameter_simple(column_name, column_values_list, operator='EQUALS', ignore_case: bool = True):
    return {
        "type": "COLUMN",
        "name": column_name,
        "values": column_values_list,
        "operator": operator,
        "ignoreCase": ignore_case
    }


def generate_policy_body(policy_name, dataset_id, parameters_list, policy_id=None, user_ids=None,
                         group_ids=None, virtual_user_ids=None):
    if not user_ids:
        user_ids = []

    if not group_ids:
        group_ids = []
    
    if not virtual_user_ids:
        virtual_user_ids = []

    body = {
        "name": policy_name,
        "dataSourceId": dataset_id,
        "userIds": user_ids,
        "virtualUserIds": virtual_user_ids,
        "groupIds": group_ids,
        "dataSourcePermissions": False,
        "parameters": parameters_list
    }

    if policy_id:
        body.update({'filterGroupId': policy_id})

    return body


async def create_policy(full_auth: DomoFullAuth, dataset_id: str, body: dict, debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/data-control/{dataset_id}/filter-groups'
    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='POST',
        body=body,
        debug=debug)

    return res


async def update_policy(full_auth: DomoFullAuth, dataset_id: str, filter_group_id, body: dict,
                        debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/data-control/{dataset_id}/filter-groups/{filter_group_id}'
    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body=body,
        debug=debug
    )

    return res




import aiohttp

from ...utils.ResponseGetData import ResponseGetData
from ..DomoAuth import DomoFullAuth
from .get_data import get_data


async def get_bootstrap(full_auth: DomoFullAuth, debug: bool, session: aiohttp.ClientSession = None) -> ResponseGetData:
    """get bootstrap data"""

    url = f'https://{full_auth.domo_instance}.domo.com/api/domoweb/bootstrap?v2Navigation=false'

    res = await get_data(url=url, method='GET', auth=full_auth, debug=debug, session=session)

    if debug:
        res.print(is_pretty=True)

    return res


async def bsr_features(full_auth: DomoFullAuth, session: aiohttp.ClientSession = None, debug: bool = False) -> list[
        dict]:
    res = await get_bootstrap(full_auth=full_auth, session=session, debug=debug)

    if res.status == 200:
        bsr = res.response.get('data').get('features')
        return bsr



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoBootstrap.ipynb.

# %% auto 0
__all__ = ['DomoBootstrap_Feature', 'DomoBootstrap']

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 2
from dataclasses import dataclass, field

import httpx

import domolibrary.classes.DomoPage as dmpg

import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.Logger as lc
import domolibrary.routes.bootstrap as bootstrap_routes


# %% ../../nbs/classes/50_DomoBootstrap.ipynb 3
from fastcore.basics import patch_to



# %% ../../nbs/classes/50_DomoBootstrap.ipynb 4
@dataclass
class DomoBootstrap_Feature:
    id: int
    name: str
    label: str
    type: str
    purchased: bool
    enabled: bool

    @classmethod
    def _from_json_bootstrap(cls, json_obj: dict):
        dd = util_dd.DictDot(json_obj)

        bsf = cls(
            id=dd.id,
            name=dd.name,
            label=dd.label,
            type=dd.type,
            purchased=dd.purchased,
            enabled=dd.enabled
        )
        return bsf

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 5
@dataclass
class DomoBootstrap:
    auth : dmda.DomoAuth = field(repr = False)
    page_ls : list[dmpg.DomoPage] = field(default = None)
    feature_ls : list[DomoBootstrap_Feature] = field(default = None)
    

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 6
@patch_to(DomoBootstrap)
async def get_all(self: DomoBootstrap,
                  auth: dmda.DomoAuth = None, debug_api: bool = False):
    
    auth = auth or self.auth

    res =  await bootstrap_routes.get_bootstrap(auth=auth, debug_api=debug_api)

    return res.response


# %% ../../nbs/classes/50_DomoBootstrap.ipynb 9
@patch_to(DomoBootstrap)
async def get_pages(self: DomoBootstrap,
                    auth: dmda.DomoAuth = None, return_raw : bool = False, debug_api: bool = False) -> list[dmpg.DomoPage]:

    auth = auth or self.auth

    res = await bootstrap_routes.get_bootstrap_pages(auth=auth,
                                                     debug_api=debug_api)
    
    if return_raw:
        return res.response
        
    if not res.is_success:
        return None
    
    page_ls = res.response

    self.page_ls =  [dmpg.DomoPage._from_bootstrap(page_obj, auth = auth) for page_obj in page_ls]

    return self.page_ls


# %% ../../nbs/classes/50_DomoBootstrap.ipynb 12
@patch_to(DomoBootstrap)
async def get_features(self : DomoBootstrap,
                        auth: dmda.DomoAuth = None,
                        debug_api: bool = False,
                        return_raw:bool = False,
                        session: httpx.AsyncClient = None, 
                        ):
    
    auth = auth or self.auth

    res = await bootstrap_routes.get_bootstrap_features(auth=auth, session=session, debug_api=debug_api)

    if return_raw:
        return res

    feature_list = [DomoBootstrap_Feature._from_json_bootstrap(
        json_obj) for json_obj in res.response]

    return feature_list



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount.ipynb.

# %% auto 0
__all__ = ['DomoAccount_Config', 'DomoAccount_Config_AbstractCredential', 'DomoAccount_Config_DatasetCopy',
           'DomoAccount_Config_Governance', 'DomoAccount_Config_HighBandwidthConnector', 'AccountConfig', 'DomoAccount',
           'DomoAccount_UpdateName_Error', 'DomoAccount_CreateAccount_Error', 'DomoAccount_DeleteAccount_Error']

# %% ../../nbs/classes/50_DomoAccount.ipynb 3
from enum import Enum
from dataclasses import dataclass, field
from abc import ABC, abstractclassmethod, abstractmethod

import datetime as dt
import re


import httpx

from fastcore.basics import patch_to

import domolibrary.utils.convert as cd
import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de
import domolibrary.routes.account as account_routes


# %% ../../nbs/classes/50_DomoAccount.ipynb 5
class DomoAccount_Config(ABC):
    """DomoAccount Config abstract base class"""

    data_provider_type: str

    @classmethod
    @abstractmethod
    def _from_json(cls, obj):
        """convert accounts API response into a class object"""
        pass

    @abstractmethod
    def to_json(self):
        """convert class object into a format the accounts API expects"""
        pass

# %% ../../nbs/classes/50_DomoAccount.ipynb 6
@dataclass
class DomoAccount_Config_AbstractCredential(DomoAccount_Config):
    data_provider_type = "abstract-credential-store"
    credentials: dict

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            credentials=dd.credentials,
        )

    def to_json(self):
        return {"credentials": self.credentials}

# %% ../../nbs/classes/50_DomoAccount.ipynb 7
@dataclass
class DomoAccount_Config_DatasetCopy(DomoAccount_Config):
    domo_instance: str
    access_token: str = field(repr=False)

    data_provider_type = "dataset-copy"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(access_token=dd.accessToken, domo_instance=dd.instance)

    def to_json(self):
        return {"accessToken": self.access_token, "instance": self.domo_instance}

# %% ../../nbs/classes/50_DomoAccount.ipynb 8
@dataclass
class DomoAccount_Config_Governance(DomoAccount_Config):
    domo_instance: str
    access_token: str = field(repr=False)

    data_provider_type = "domo-governance-d14c2fef-49a8-4898-8ddd-f64998005600"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(access_token=dd.apikey, domo_instance=dd.customer)

    def to_json(self):
        return {"apikey": self.access_token, "customer": self.domo_instance}

# %% ../../nbs/classes/50_DomoAccount.ipynb 9
@dataclass
class DomoAccount_Config_HighBandwidthConnector(DomoAccount_Config):
    aws_access_key: str
    aws_secret_key: str = field(repr=False)
    s3_staging_dir: str
    region: str = "us-west-2"

    data_provider_type = "amazon-athena-high-bandwidth"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            aws_access_key=dd.awsAccessKey,
            aws_secret_key=dd.awsSecretKey,
            s3_staging_dir=dd.s3StagingDir,
            region=dd.region,
        )

    def to_json(self):
        return {
            "awsAccessKey": self.aws_access_key,
            "awsSecretKey": self.aws_secret_key,
            "s3StagingDir": self.s3_staging_dir,
            "region": self.region,
        }

# %% ../../nbs/classes/50_DomoAccount.ipynb 10
class AccountConfig(Enum):
    """
    Enum provides appropriate spelling for data_provider_type and config object.
    The name of the enum should correspond with the data_provider_type with hyphens replaced with underscores.
    """

    amazon_athena_high_bandwidth = DomoAccount_Config_HighBandwidthConnector

    abstract_credential_store = DomoAccount_Config_AbstractCredential

    dataset_copy = DomoAccount_Config_DatasetCopy

    domo_governance_d14c2fef_49a8_4898_8ddd_f64998005600 = DomoAccount_Config_Governance

# %% ../../nbs/classes/50_DomoAccount.ipynb 12
@dataclass
class DomoAccount:
    name: str
    data_provider_type: str

    id: int = None
    created_dt: dt.datetime = None
    modified_dt: dt.datetime = None
    auth: dmda.DomoAuth = field(repr=False, default=None)

    config: DomoAccount_Config = None

    @classmethod
    def _from_json(cls, obj: dict, auth: dmda.DomoAuth = None):
        """converts data_v1_accounts API response into an accounts class object"""

        dd = util_dd.DictDot(obj)

        return cls(
            id=dd.id,
            name=dd.displayName,
            data_provider_type=dd.dataProviderType,
            created_dt=cd.convert_epoch_millisecond_to_datetime(dd.createdAt),
            modified_dt=cd.convert_epoch_millisecond_to_datetime(dd.modifiedAt),
            auth=auth,
        )

    async def _get_config(
        self, session=None, debug_api: bool = None, return_raw: bool = False
    ):

        res_config = await account_routes.get_account_config(
            auth=self.auth,
            account_id=self.id,
            data_provider_type=self.data_provider_type,
            session=session,
            debug_api=debug_api,
        )

        if return_raw:
            return res_config

        enum_clean = re.sub("-", "_", self.data_provider_type)

        if not enum_clean in AccountConfig.__members__:
            return None

        self.config = AccountConfig[enum_clean].value._from_json(res_config.response)

        return self.config

# %% ../../nbs/classes/50_DomoAccount.ipynb 13
@patch_to(DomoAccount, cls_method=True)
async def get_from_id(
    cls,
    auth: dmda.DomoAuth,
    account_id: int,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    debug_api: bool = False,
):

    res = await account_routes.get_account_from_id(
        auth=auth, account_id=account_id, session=session, debug_api=debug_api
    )
    if return_raw:
        return res

    if not res.is_success:
        return None

    obj = res.response
    acc = cls._from_json(obj, auth)

    await acc._get_config(session=session, debug_api=debug_api)

    return acc

# %% ../../nbs/classes/50_DomoAccount.ipynb 17
@patch_to(DomoAccount)
async def update_config(
    self: DomoAccount,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    config: DomoAccount_Config = None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    auth = auth or self.auth

    config = config or self.config

    res = await account_routes.update_account_config(
        auth=auth,
        account_id=self.id,
        data_provider_type=self.data_provider_type,
        config_body=config.to_json(),
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    await self._get_config(session=session, debug_api=debug_api)

    return self

# %% ../../nbs/classes/50_DomoAccount.ipynb 20
class DomoAccount_UpdateName_Error(de.DomoError):
    def __init__(self,
                 domo_instance,
                 status, message,
                 entity_id,
                 function_name="update_name",
                 ):

        super().__init__(function_name=function_name, entity_id=entity_id,
                         domo_instance=domo_instance, status=status, message=message)


@patch_to(DomoAccount)
async def update_name(
    self: DomoAccount,
    account_name: str = None,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    auth = auth or self.auth

    # print(auth, self.id, self.data_provider_type, self.config.to_json())

    res = await account_routes.update_account_name(
        auth=auth,
        account_id=self.id,
        account_name=account_name or self.name,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise DomoAccount_UpdateName_Error(entity_id=self.id,
                                           domo_instance=auth.domo_instance,
                                           status=res.status,
                                           message=res.response,)

    self = await self.get_from_id(auth=auth, account_id=self.id)

    return self


# %% ../../nbs/classes/50_DomoAccount.ipynb 24
class DomoAccount_CreateAccount_Error(de.DomoError):
    def __init__(self,
                entity_id,
                 domo_instance,
                 status, 
                 message,
                 function_name="create_account",
                 ):

        super().__init__(function_name=function_name, entity_id = entity_id,
                         domo_instance=domo_instance, status=status, message=message)


@patch_to(DomoAccount, cls_method=True)
def generate_create_body(cls, account_name, config):
    return {
        "displayName": account_name,
        "dataProviderType": config.data_provider_type,
        "name": config.data_provider_type,
        "configurations": config.to_json(),
    }


@patch_to(DomoAccount, cls_method=True)
async def create_account(
    cls: DomoAccount,
    account_name: str,
    config: DomoAccount_Config,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):

    body = cls.generate_create_body(account_name=account_name, config=config)

    res = await account_routes.create_account(
        auth=auth, config_body=body, debug_api=debug_api, session=session
    )

    if not res.is_success:
        raise DomoAccount_CreateAccount_Error(
            entity_id=account_name,
            domo_instance=auth.domo_instance,
            status=res.status,
            message=res.response,
        )

    return await cls.get_from_id(auth=auth, account_id=res.response.get("id"))


# %% ../../nbs/classes/50_DomoAccount.ipynb 25
class DomoAccount_DeleteAccount_Error(de.DomoError):
    def __init__(self,
                entity_id,
                 domo_instance,
                 status, message,
                 function_name="delete_account",
                 ):

        super().__init__(function_name=function_name, entity_id = entity_id,
                         domo_instance=domo_instance, status=status, message=message)


@patch_to(DomoAccount)
async def delete_account(
    self: DomoAccount,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):

    auth = auth or self.auth

    res = await account_routes.delete_account(
        auth=auth, account_id=self.id, debug_api=debug_api, session=session
    )

    if not res.is_success:
        
        raise DomoAccount_DeleteAccount_Error(
            entity_id=self.id,
            domo_instance=auth.domo_instance,
            status=res.status,
            message=res.response,
        )

    return True



import json
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any


class Config(Enum):
    PGSQL = {
        "type": "CONNECTOR",
        "description": "com.domo.connector.postgre.data",
        "version": "2",
        "dataProvider_key": "postgresql"
    }
    PGSQL_PARTITION = {
        "type": "CONNECTOR",
        "description": "com.domo.connector.postgresql.partition",
        "version": "0",
        "dataProvider_key": "postgresql-partition"
    }

    ATHENA_HIGHBANDWIDTH = {
        "type": "CONNECTOR",
        "description": "com.domo.connector.amazonathena.highbandwidth",
        "version": "0",
        "dataProvider_key": "amazon-athena-high-bandwidth"
    }

    SNOWFLAKE_UNLOADv2 = {
        "type": "CONNECTOR",
        "description": "com.domo.connector.snowflakeunloadv2",
        "version": "1",
        "dataProvider_key": "snowflake-unload-v2"
    }


class UpdateMethod(Enum):
    REPLACE = 'REPLACE'
    APPEND = 'APPEND'


@dataclass
class Transport:
    type: str
    description: str
    version: str


class ConfigurationParam_Category(Enum):
    METADATA = 'METADATA'


@dataclass
class ConfigurationParam:
    name: str
    value: Any
    type: str = 'string'
    category: ConfigurationParam_Category = ConfigurationParam_Category.METADATA.value


@dataclass
class Account:
    id: int


@dataclass
class Datasource:
    name: str
    description: str


class Schedule_Type(Enum):
    DAY = 'DAY'
    MANUAL = 'MANUAL'


@dataclass
class Schedule:
    type: Schedule_Type
    at: str = None
    timezone: str = 'UTC'

    def to_dict(self):
        return asdict(self)


@dataclass
class DataProvider:
    key: str


@dataclass
class Configuration:
    transport: Transport
    configuration: list[ConfigurationParam]
    account: Account
    updateMethod: UpdateMethod
    dataProvider: DataProvider
    dataSource: Datasource
    advancedScheduleJson: Schedule

    def __post_init__(self):
        if not self.advancedScheduleJson:
            self.advancedScheduleJson = Schedule(
                type=Schedule_Type.MANUAL.value)

    def to_dict(self):
        clean_self = self

        advancedSchedule = clean_self.advancedScheduleJson.to_dict()

        if advancedSchedule.get('at') is None:
            del advancedSchedule['at']

        clean_self.advancedScheduleJson = json.dumps(advancedSchedule)
        return asdict(clean_self)



import csv
import datetime as DT
import os
from dataclasses import dataclass

import aiohttp
import Library.DomoClasses.DomoAuth as dmda
import Library.DomoClasses.DomoDataset as dmds
import Library.DomoClasses.DomoInstanceConfig as dmic
import Library.utils.Exceptions as ex
import pandas as pd


@dataclass
class LogError:
    def __init__(self):
        pass

    function_str: str
    message_str: str
    domo_instance: str


def write_error(file_path, log_err: LogError):
    file_exists = os.path.isfile(file_path)

    with open(file_path, 'a+') as log_file:
        headers = list(log_err.__dict__.keys())
        writer = csv.DictWriter(log_file, fieldnames=headers)

        if not file_exists:
            writer.writeheader()

        writer.writerows([log_err.__dict__])

        # res = log_file.write(f"{log_err.domo_instance}, {log_err.function_str}, {log_err.message_str}\n")


async def get_ip_whitelist_config(config_full_auth: dmda.DomoFullAuth,
                                  dataset_id: str,
                                  handle_err_fn: callable,
                                  sql: str = "select addresses from table",
                                  debug: bool = False):
    try:
        sync_ip_ds = await dmds.DomoDataset.get_from_id(full_auth=config_full_auth,
                                                        id=dataset_id,
                                                        debug=debug)
        if debug:
            print(sync_ip_ds)

        print(
            f"⚙️ START - Retrieving whitelist configuration \n{sync_ip_ds.display_url()}")

        sync_ip_df = await sync_ip_ds.query_dataset_private(full_auth=config_full_auth,
                                                            dataset_id=dataset_id,
                                                            sql=sql)

        if sync_ip_df.empty:
            raise Exception('no whitelist returned')
            return False

        print(
            f"\n⚙️ SUCCESS 🎉 Retrieved whitelist configuration  \nThere are {len(sync_ip_df.index)} ip addresses to sync")

        return list(sync_ip_df['addresses'])

    except ex.InvalidDataset:
        print('invalid dataset')

        handle_err_fn(log_err=LogError(function_str='get_ip_whitelist_config',
                                       message_str=f'invalid dataset {dataset_id} not matched in {config_full_auth.domo_instance}',
                                       domo_instance=config_full_auth.domo_instance))
        return False

    except Exception:
        print("did it fail?")
        handle_err_fn(log_err=LogError(function_str='get_ip_whitelist_config',
                                       message_str=f'undefined error',
                                       domo_instance=config_full_auth.domo_instance))
        return False


async def remove_partition_by_x_days(full_auth: dmda.DomoFullAuth,
                                     dataset_id: str,
                                     x_last_days: int = 0,
                                     separator: str = None,
                                     date_index: int = 0,
                                     date_format: str = '%Y-%m-%d',
                                     session: aiohttp.ClientSession = None
                                     ):
    domo_ds = dmds.DomoDataset(full_auth=full_auth, id=dataset_id)

    list_partition = await domo_ds.list_partitions(full_auth=full_auth, dataset_id=dataset_id)

    today = DT.date.today()
    days_ago = today - DT.timedelta(days=x_last_days)
    for i in list_partition:
        compare_date = ''
        if separator is not None and separator != '':
            compare_date = i['partitionId'].split(separator)[date_index]
        else:
            compare_date = i['partitionId']

        try:
            d = DT.datetime.strptime(compare_date, date_format).date()
        except ValueError:
            d = None
        if d is not None and d < days_ago:
            print(full_auth.domo_instance, ': 🚀  Removing partition key : ',
                  (i['partitionId']), ' in ', dataset_id)
            await domo_ds.delete_partition(dataset_partition_id=i['partitionId'], dataset_id=dataset_id,
                                           full_auth=full_auth, session=session)


async def get_instance_whitelist_df(instance_auth: dmda.DomoFullAuth,
                                    session: aiohttp.ClientSession = None,
                                    debug: bool = False) -> pd.DataFrame:
    """return a dataframe data in the correct shape for upload for ONE instance"""
    instance_whitelist = await dmic.DomoInstanceConfig.get_whitelist(full_auth=instance_auth, session=session)

    if instance_whitelist == ['']:
        instance_whitelist = ['no_ip_whitelist']

    upload_df = pd.DataFrame(instance_whitelist, columns=['address'])
    upload_df['instance'] = instance_auth.domo_instance
    upload_df['url'] = f'https://{instance_auth.domo_instance}.domo.com/admin/security/whitelist'

    return upload_df


async def get_company_domains(config_full_auth: dmda.DomoFullAuth,
                              dataset_id: str,
                              handle_err_fn: callable,
                              sql: str = "select domain from table",
                              global_admin_username: str = None,
                              global_admin_password: str = None,
                              execution_env: str = None,
                              debug: bool = False) -> pd.DataFrame:
    ds = await dmds.DomoDataset.get_from_id(full_auth=config_full_auth,
                                            id=dataset_id, debug=debug)

    print(f"⚙️ START - Retrieving company list \n{ds.display_url()}")
    print(f"⚙️ SQL = {sql}")

    df = await ds.query_dataset_private(full_auth=config_full_auth,
                                        dataset_id=dataset_id,
                                        sql=sql,
                                        debug=debug)

    df["domo_instance"] = df["domain"].apply(
        lambda x: x.replace('.domo.com', ''))

    if global_admin_username:
        df["domo_username"] = global_admin_username
    if global_admin_password:
        df["domo_password"] = global_admin_password

    if execution_env:
        df['env'] = execution_env or 'manual'

    if df.empty:
        raise Exception('no companies retrieved')
        return False

    print(
        f"\n⚙️ SUCCESS 🎉 Retrieved company list \nThere are {len(df.index)} companies to update")

    return df



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoDataset.ipynb.

# %% auto 0
__all__ = ['DatasetSchema_Types', 'DomoDataset_Schema_Column', 'DomoDataset_Schema', 'DatasetTags_SetTagsError',
           'DomoDataset_Tags', 'DomoDataset', 'QueryExecutionError', 'DomoDataset_DeleteDataset_Error',
           'DomoDataset_UploadData_Error', 'DomoDataset_UploadData_DatasetUploadId_Error',
           'DomoDataset_UploadData_UploadData_Error', 'DomoDataset_CreateDataset_Error']

# %% ../../nbs/classes/50_DomoDataset.ipynb 4
from fastcore.basics import patch_to
import pandas as pd

# %% ../../nbs/classes/50_DomoDataset.ipynb 5
from dataclasses import dataclass, field
from typing import Any, List, Optional
from enum import Enum, auto

import json
import io

import httpx
import asyncio


# import datetime as dt

# import importlib
# import json
# from pprint import pprint

# import pandas as pd


# from ..utils.Base import Base
# from ..utils.chunk_execution import chunk_list
# from . import DomoCertification as dmdc
# from . import DomoPDP as dmpdp
# from . import DomoTag as dmtg


import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de
import domolibrary.routes.dataset as dataset_routes

# %% ../../nbs/classes/50_DomoDataset.ipynb 7
async def _have_prereqs(self, auth, dataset_id, function_name):
    """tests if have a parent dataset or prerequsite dataset_id and auth object"""

    auth_from_self_dataset = getattr(self.dataset, 'auth', None) if getattr(self, 'dataset', None) else None
    auth_from_self = getattr(self , 'auth', None)

    auth = auth or auth_from_self or auth_from_self_dataset

    await auth.get_auth_token()

    if not auth or not auth.token:
        raise de.AuthNotProvidedError(
            function_name=function_name,
            entity_id = self.dataset.id)

    id_from_self = getattr(self, 'id', None)
    id_from_self_parent = getattr(self.dataset, 'id', None ) if getattr(self, 'dataset', None) else None
    
    dataset_id = dataset_id or id_from_self or id_from_self_parent
    
    if not dataset_id:
        raise de.DatasetNotProvidedError(
            function_name = function_name, 
            domo_instance = auth.domo_instance
        )

    return auth, dataset_id

# %% ../../nbs/classes/50_DomoDataset.ipynb 8
class DatasetSchema_Types(Enum):
    STRING = 'STRING'
    DOUBLE = 'DOUBLE'
    LONG = 'LONG'
    DATE = 'DATE'
    DATETIME = 'DATETIME'


@dataclass
class DomoDataset_Schema_Column:
    name: str
    id: str
    type: DatasetSchema_Types

    @classmethod
    def _from_json(cls, json_obj):
        dd = util_dd.DictDot(json_obj)
        return cls(name=dd.name, id=dd.id, type=dd.type)


@dataclass
class DomoDataset_Schema:
    """class for interacting with dataset schemas"""

    dataset: any = None
    columns: List[DomoDataset_Schema_Column] = field(default_factory=list)

    async def get(
        self,
        auth: Optional[dmda.DomoAuth] = None,
        dataset_id: str = None,
        debug_api: bool = False,
        return_raw_res: bool = False,  # return the raw response
    ) -> List[DomoDataset_Schema_Column]:

        """method that retrieves schema for a dataset"""

        auth, dataset_id = await _have_prereqs(self = self, auth = auth, dataset_id = dataset_id, function_name = "DomoDataset_Schema.get")

        res = await dataset_routes.get_schema(
            auth=auth, dataset_id=dataset_id, debug_api=debug_api
        )

        if return_raw_res:
            return res.response

        if res.status == 200:
            json_list = res.response.get("tables")[0].get("columns")

            self.columns = [
                DomoDataset_Schema_Column._from_json(json_obj=json_obj)
                for json_obj in json_list
            ]

            return self.columns


# %% ../../nbs/classes/50_DomoDataset.ipynb 13
class DatasetTags_SetTagsError(Exception):
    """return if DatasetTags request is not successfull"""

    def __init__(self, dataset_id, domo_instance):
        message = f"failed to set tags on dataset - {dataset_id} in {domo_instance}"
        super().__init__(message)


@dataclass
class DomoDataset_Tags:
    """class for interacting with dataset tags"""

    dataset: any = None
    tag_ls: List[str] = field(default_factory=list)

    async def get(
        self,
        dataset_id: str = None,
        auth: Optional[dmda.DomoAuth] = None,
        debug_api: bool = False,
        session: Optional[httpx.AsyncClient] = None,
    ) -> List[str]:  # returns a list of tags
        """gets the existing list of dataset_tags"""

        auth, dataset_id = await _have_prereqs(self = self, auth=auth, dataset_id=dataset_id, function_name="DomoDataset_Tages.get")

        res = await dataset_routes.get_dataset_by_id(
            dataset_id=dataset_id, auth=auth, debug_api=debug_api, session=session
        )

        if res.is_success == False:
            print(res)
            return None

        if res.is_success == True:
            tag_ls = json.loads(res.response.get("tags"))
            self.tag_ls = tag_ls

            return tag_ls

    async def set(
        self,
        tag_ls: [str],
        dataset_id: str = None,
        auth: Optional[dmda.DomoAuth] = None,
        debug_api: bool = False,
        session: Optional[httpx.AsyncClient] = None,
    ) -> List[str]: # returns a list of tags
        """replaces all tags with a new list of dataset_tags"""

        auth, dataset_id = await _have_prereqs(self = self , auth=auth, dataset_id=dataset_id, function_name="DomoDatasetTags.set")

        res = await dataset_routes.set_dataset_tags(
            auth=auth,
            tag_ls=list(set(tag_ls)),
            dataset_id=dataset_id,
            debug_api=debug_api,
            session=session,
        )

        if res.status != 200:
            raise DatasetTags_SetTagsError(
                dataset_id=dataset_id, domo_instance=auth.domo_instance
            )

        await self.get(dataset_id=dataset_id, auth=auth)

        return self.tag_ls

# %% ../../nbs/classes/50_DomoDataset.ipynb 16
@patch_to(DomoDataset_Tags)
async def add(
    self: DomoDataset_Tags,
    add_tag_ls: [str],
    dataset_id: str = None,
    auth: Optional[dmda.DomoAuth] = None,
    debug_api: bool = False,
    session: Optional[httpx.AsyncClient] = None,
) -> List[str]:  # returns a list of tags
    """appends tags to the list of existing dataset_tags"""

    auth, dataset_id = await _have_prereqs(self = self, auth=auth, dataset_id=dataset_id, function_name = "DomoDataset_Tags.add")

    existing_tag_ls = await self.get(dataset_id=dataset_id, auth=auth)
    
    add_tag_ls += existing_tag_ls

    return await self.set(
        auth=auth,
        dataset_id=dataset_id,
        tag_ls=list(set(add_tag_ls)),
        debug_api=debug_api,
        session=session,
    )

# %% ../../nbs/classes/50_DomoDataset.ipynb 18
@patch_to(DomoDataset_Tags)
async def remove(self: DomoDataset_Tags,
                 remove_tag_ls: [str],
                 dataset_id: str = None,
                 auth: dmda.DomoFullAuth = None,
                 debug_api: bool = False,
                 session: Optional[httpx.AsyncClient] = None
                 ) -> List[str]:  # returns a list of tags
    """removes tags from the existing list of dataset_tags"""

    auth, dataset_id = await _have_prereqs(self = self, auth=auth, dataset_id=dataset_id, function_name = "DomoDataset_Tags.remove")

    existing_tag_ls = await self.get(dataset_id=dataset_id, auth=auth)

    existing_tag_ls = [
        ex for ex in existing_tag_ls if ex not in remove_tag_ls]

    return await self.set(auth=auth,
                          dataset_id=dataset_id,
                          tag_ls=list(set(existing_tag_ls)),
                          debug_api=debug_api, session=session)


# %% ../../nbs/classes/50_DomoDataset.ipynb 22
@dataclass
class DomoDataset:
    "interacts with domo datasets"

    auth: dmda.DomoAuth = field(repr=False, default=None)

    id: str = ""
    display_type: str = ""
    data_provider_type: str = ""
    name: str = ""
    description: str = ""
    row_count: int = None
    column_count: int = None

    stream_id: int = None

    owner: dict = field(default_factory=dict)
    formula: dict = field(default_factory=dict)

    schema: DomoDataset_Schema = field(default=None)
    # tags: Dataset_Tags = field(default = None)

    # certification: dmdc.DomoCertification = None
    # PDPPolicies: dmpdp.Dataset_PDP_Policies = None

    def __post_init__(self):
        self.schema = DomoDataset_Schema(dataset=self)
        # self.tags = Dataset_Tags(dataset=self)

        # self.PDPPolicies = dmpdp.Dataset_PDP_Policies(self)

    def display_url(self):
        return f"https://{self.auth.domo_instance }.domo.com/datasources/{self.id}/details/overview"

# %% ../../nbs/classes/50_DomoDataset.ipynb 26
@patch_to(DomoDataset, cls_method=True)
async def get_from_id(
    cls: DomoDataset,
    dataset_id: str,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    return_raw_res: bool = False,
):

    """retrieves dataset metadata"""

    res = await dataset_routes.get_dataset_by_id(
        auth=auth, dataset_id=dataset_id, debug_api=debug_api
    )

    if return_raw_res:
        return res.response

    dd = util_dd.DictDot(res.response)
    ds = cls(
        auth=auth,
        id=dd.id,
        display_type=dd.displayType,
        data_provider_type=dd.dataProviderType,
        name=dd.name,
        description=dd.description,
        owner=dd.owner,
        formula=dd.properties.formulas.formulas,
        stream_id=dd.streamId,
        row_count=int(dd.rowCount),
        column_count=int(dd.columnCount),
    )

    # if dd.tags:
    #     ds.tags.tag_ls = json.loads(dd.tags)

    # if dd.certification:
    #     # print('class def certification', dd.certification)
    #     ds.certification = dmdc.DomoCertification._from_json(
    #         dd.certification)

    return ds


# %% ../../nbs/classes/50_DomoDataset.ipynb 30
class QueryExecutionError(de.DomoError):
    def __init__(self,
                 sql, dataset_id,
                 domo_instance,
                 status, message,
                 function_name=None ):
        
        self.message = f"error executing {sql}: {message}"

        super().__init__(entity_id=dataset_id,
                         function_name=function_name,
                         status=status,
                         message=message,
                         domo_instance=domo_instance)


@patch_to(DomoDataset, cls_method=True)
async def query_dataset_private(cls: DomoDataset,
                                auth: dmda.DomoAuth,  # DomoFullAuth or DomoTokenAuth
                                dataset_id: str,
                                sql: str,
                                session: Optional[httpx.AsyncClient] = None,
                                loop_until_end: bool = False,  # retrieve all available rows
                                limit=100,  # maximum rows to return per request.  refers to PAGINATION
                                skip=0,
                                maximum=100,  # equivalent to the LIMIT or TOP clause in SQL, the number of rows to return total
                                debug_api: bool = False,
                                debug_loop: bool = False,
                                ) -> pd.DataFrame:

    res = await dataset_routes.query_dataset_private(auth=auth,
                                                     dataset_id=dataset_id,
                                                     sql=sql,
                                                     maximum=maximum,
                                                     skip=skip,
                                                     limit=limit,
                                                     loop_until_end=loop_until_end,
                                                     session=session,
                                                     debug_loop=debug_loop,
                                                     debug_api=debug_api
                                                     )

    if not res.is_success:
        raise QueryExecutionError(
            status=res.status, message=res.response,
            function_name="query_dataset_private", 
            sql=sql, dataset_id=dataset_id, domo_instance=auth.domo_instance)

    return pd.DataFrame(res.response)


# %% ../../nbs/classes/50_DomoDataset.ipynb 31
class DomoDataset_DeleteDataset_Error(de.DomoError):
    def __init__(self,
                 dataset_id,
                 status, reason,
                 domo_instance,
                 function_name
                 ):

        super().__init__(entity_id=dataset_id,
                         function_name=function_name,
                         status=status,
                         message=reason,
                         domo_instance=domo_instance)


@patch_to(DomoDataset)
async def delete(self: DomoDataset,
                 dataset_id=None,
                 auth: dmda.DomoAuth = None,
                 debug_api: bool = False,
                 session: httpx.AsyncClient = None):

    dataset_id = dataset_id or self.id
    auth = auth or self.auth

    res = await dataset_routes.delete(
        auth=auth,
        dataset_id=dataset_id,
        debug_api=debug_api,
        session=session)

    if not res.is_success:
        raise DomoDataset_DeleteDataset_Error(
            dataset_id=dataset_id, 
            function_name="DomoDataset.delete",
            domo_instance=auth.domo_instance, 
            status=res.status, reason=res.response)

    return res


# %% ../../nbs/classes/50_DomoDataset.ipynb 34
class DomoDataset_UploadData_Error(Exception):

    def __init__(self,
                 message_error: str,
                 domo_instance: str,
                 dataset_id: str,
                 stage: int,
                 status="", reason="",
                 partition_key: str = None):

        message_start = f"Stage {stage}:: {message_error} :: API {status} - {reason} :: "
        message_end = f"in {dataset_id} in {domo_instance}"

        message_partition = ""
        if partition_key:
            message_partition = f"for partition - '{partition_key}' "

        message = f"{message_start}{message_partition}{message_end}"

        super().__init__(message)


class DomoDataset_UploadData_DatasetUploadId_Error(DomoDataset_UploadData_Error):
    def __init__(self, domo_instance: str, dataset_id: str,
                 stage: int = 1, status="", reason="",
                 partition_key: str = None):

        message_error = "unable to retrieve dataset_upload_id"

        super().__init__(message_error=message_error,
                         domo_instance=domo_instance, dataset_id=dataset_id,
                         stage=stage, status=status, reason=reason,
                         partition_key=partition_key)


class DomoDataset_UploadData_UploadData_Error(DomoDataset_UploadData_Error):
    def __init__(self, domo_instance: str, dataset_id: str,
                 stage: int = 2, status="", reason="",
                 partition_key: str = None):

        message_error = "while uploading data"

        super().__init__(message_error=message_error,
                         domo_instance=domo_instance, dataset_id=dataset_id,
                         stage=stage, status=status, reason=reason,
                         partition_key=partition_key)

    class DomoDataset_UploadData_CommitDatasetUploadId_Error(DomoDataset_UploadData_Error):
        def __init__(self, domo_instance: str, dataset_id: str,
                     stage: int = 3, status="", reason="",
                     partition_key: str = None):

            message_error = "while commiting dataset_upload_id"

            super().__init__(message_error=message_error,
                              domo_instance=domo_instance, dataset_id=dataset_id,
                              stage=stage, status=status, reason=reason,
                              partition_key=partition_key)


# %% ../../nbs/classes/50_DomoDataset.ipynb 35
@patch_to(DomoDataset)
async def index_dataset(self: DomoDataset,
                        auth: dmda.DomoAuth = None,
                        dataset_id: str = None,
                        debug_api: bool = False,
                        session: httpx.AsyncClient = None
                        ):

    auth = auth or self.auth
    dataset_id = dataset_id or self.id
    return await dataset_routes.index_dataset(auth=auth, dataset_id=dataset_id, debug_api=debug_api,
                                              session=session)


# %% ../../nbs/classes/50_DomoDataset.ipynb 36
@patch_to(DomoDataset)
async def upload_data(self : DomoDataset,
                      upload_df: pd.DataFrame = None,
                      upload_df_ls: list[pd.DataFrame] = None,
                      upload_file: io.TextIOWrapper = None,

                      upload_method: str = 'REPLACE',  # APPEND or REPLACE
                      partition_key: str = None,

                      is_index: bool = True,

                      dataset_id: str = None,
                      dataset_upload_id=None,

                      auth: dmda.DomoAuth = None,

                      session: httpx.AsyncClient = None,
                      debug_api: bool = False,
                      debug_prn: bool = False
                      ):

    auth, dataset_id = await _have_prereqs(self = self, auth = auth, dataset_id=dataset_id, function_name= "upload_data")

    upload_df_ls = upload_df_ls or [upload_df]

    status_message = f"{dataset_id} {partition_key} | {auth.domo_instance}"

    # stage 1 get uploadId
    if not dataset_upload_id:
        if debug_prn:
            print(f"\n\n🎭 starting Stage 1 - {status_message}")

        stage_1_res = await dataset_routes.upload_dataset_stage_1(auth=auth,
                                                                  dataset_id=dataset_id,
                                                                  session=session,
                                                                  partition_tag=partition_key,
                                                                  debug_api=debug_api
                                                                  )
        if debug_prn:
            print(f"\n\n🎭 Stage 1 response -- {stage_1_res.status} for {status_message}")

        dataset_upload_id = stage_1_res.response.get('uploadId')

    if not dataset_upload_id:
        raise DomoDataset_UploadData_DatasetUploadId_Error(
            domo_instance=auth.domo_instance,  dataset_id=dataset_id, stage=1, partition_key=partition_key,
            status=stage_1_res.status, reason=stage_1_res.response)

    # stage 2 upload_dataset
    stage_2_res = None

    if upload_file:
        if debug_prn:
            print(f"\n\n🎭 starting Stage 2 - upload file for {status_message}")

        stage_2_res = await asyncio.gather(*[dataset_routes.upload_dataset_stage_2_file(auth=auth,
                                                                                        dataset_id=dataset_id,
                                                                                        upload_id=dataset_upload_id,
                                                                                        part_id=1,
                                                                                        data_file=upload_file,
                                                                                        session=session, debug_api=debug_api)])

    else:
        if debug_prn:
            print(
                f"\n\n🎭 starting Stage 2 - {len(upload_df_ls)} - number of parts for {status_message}")
        stage_2_res = await asyncio.gather(*[dataset_routes.upload_dataset_stage_2_df(auth=auth,
                                                                                      dataset_id=dataset_id,
                                                                                      upload_id=dataset_upload_id,
                                                                                      part_id=index + 1,
                                                                                      upload_df=df,
                                                                                      session=session, debug_api=debug_api) for index, df in enumerate(upload_df_ls)])

    for res in stage_2_res:
        if not res.is_success:
            raise DomoDataset_UploadData_UploadData_Error(
                domo_instance=auth.domo_instance, dataset_id=dataset_id, stage=2, partition_key=partition_key,
                status=res.status, reason=res.response)

    if debug_prn:
        print(f"🎭 Stage 2 - upload data: complete for {status_message}")

    # stage 3 commit_data
    if debug_prn:
        print(f"\n\n🎭 starting Stage 3 - commit dataset_upload_id for {status_message}")

    await asyncio.sleep(10)  # wait for uploads to finish
    stage3_res = await dataset_routes.upload_dataset_stage_3(auth=auth,
                                                             dataset_id=dataset_id,
                                                             upload_id=dataset_upload_id,
                                                             update_method=upload_method,
                                                             partition_tag=partition_key,
                                                             is_index=False,
                                                             session=session,
                                                             debug_api=debug_api)

    if not stage3_res.is_success:
        raise DomoDataset_UploadData_CommitDatasetUploadId_Error(
            domo_instance=auth.domo_instance, dataset_id=dataset_id, partition_key=partition_key, stage=3,
            status=stage3_res.status, reason=stage3_res.response)

    if debug_prn:
        print(f"\n🎭 stage 3 - commit dataset: complete for {status_message} ")

    if is_index:
        await asyncio.sleep(3)
        return await self.index_dataset(auth=auth,
                                        dataset_id=dataset_id,
                                        debug_api=debug_api,
                                        session=session)

    return stage3_res


# %% ../../nbs/classes/50_DomoDataset.ipynb 38
@patch_to(DomoDataset)
async def list_partitions(self : DomoDataset,
                            auth: dmda.DomoAuth = None,
                            dataset_id: str = None,
                            debug_api: bool = False,
                            session: httpx.AsyncClient = None
                            ):

    auth = auth or self.auth
    dataset_id = dataset_id or self.id

    res = await dataset_routes.list_partitions(auth=auth, dataset_id=dataset_id, debug_api=debug_api,
                                                session=session)
    if res.status != 200:
        return None

    return res.response

# %% ../../nbs/classes/50_DomoDataset.ipynb 40
class DomoDataset_CreateDataset_Error(Exception):
    def __init__(self, domo_instance: str, dataset_name: str, status: int, reason: str):
        message = f"Failure to create dataset {dataset_name} in {domo_instance} :: {status} - {reason}"


@patch_to(DomoDataset, cls_method=True)
async def create(cls: DomoDataset,
                 dataset_name: str,
                 dataset_type='api',

                 schema={"columns": [
                     {"name": 'col1', "type": 'LONG', "upsertKey": False},
                     {"name": 'col2', "type": 'STRING', "upsertKey": False}
                 ]},
                 auth: dmda.DomoAuth = None,
                 debug_api: bool = False, 
                 session : httpx.AsyncClient = None
                 ):
    

    res = await dataset_routes.create(dataset_name=dataset_name,
                                      dataset_type=dataset_type,
                                      schema=schema, auth=auth, debug_api=debug_api, session=session
                                      )

    if not res.is_success:
        raise DomoDataset_CreateDataset_Error(
            domo_instance=auth.domo_instance, dataset_name=dataset_name, 
            status=res.status, reason=res.response)

    dataset_id = res.response.get('dataSource').get('dataSourceId')

    return await cls.get_from_id(dataset_id=dataset_id, auth=auth)




import aiohttp

import Library.DomoClasses.DomoAuth as dmda
import Library.utils.ResponseGetData as rgd

from Library.DomoClasses.routes.get_data import get_data


async def search_publications(full_auth: dmda.DomoFullAuth, search_term: str = None, limit=100, offset=0, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:
    url = f"https://{full_auth.domo_instance}.domo.com/api/publish/v2/publication/summaries"

    params = {'limit': limit, 'offset': offset}

    if search_term:
        params.update({'searchTerm': search_term})

    res = await get_data(auth=full_auth,
                         method='GET',
                         url=url,
                         params=params,
                         session=session,
                         debug=debug)

    return res


async def get_publication_by_id(full_auth: dmda.DomoFullAuth, publication_id: str, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:
    url = f"https://{full_auth.domo_instance}.domo.com/api/publish/v2/publication/{publication_id}"

    res = await get_data(auth=full_auth,
                         method='GET',
                         url=url,
                         session=session,
                         debug=debug)

    return res


# generate publish body
def generate_publish_body(url: str, sub_domain_ls: [str], content_ls: [str], name: str, description: str, unique_id: str, is_new: bool):
    if not sub_domain_ls:
        sub_domain_ls = []

    if not content_ls:
        content_ls = []

    body = {
        "id": unique_id,
        "name": name,
        "description": description,
        "domain": url,
        "content": content_ls,
        "subscriberDomain": sub_domain_ls,
        "new": str(is_new).lower()
    }

    return body

# Creating publish job for a specific subscriber

async def create_publish_job(full_auth: dmda.DomoFullAuth, body : dict, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/publish/v2/publication'

    print (body)
    res = await get_data(auth=full_auth,
                         method='POST',
                         url=url,
                         body=body,
                         session=session,
                         debug=debug)

    return res

# Updating existing publish job with content

async def udpate_publish_job(full_auth: dmda.DomoFullAuth, publication_id: str, body: dict, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:

    url = f'https://{full_auth.domo_instance}.domo.com/api/publish/v2/publication/{publication_id}'
    
    res = await get_data(auth=full_auth,
                         method='PUT',
                         url=url,
                         body=body,
                         session=session,
                         debug=debug)
    return res

# finds all jobs waiting to be accepted within the subscriber

async def get_subscription_invites_list(full_auth: dmda.DomoFullAuth, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:

    url = f'https://{full_auth.domo_instance}.domo.com/api/publish/v2/subscription/invites'

    res = await get_data(auth=full_auth,
                         method='GET',
                         url=url,
                         session=session,
                         debug=debug)
    return res

# this takes get_subscription_invites_list into account and accepts - not instant


async def accept_invite_by_id(full_auth: dmda.DomoFullAuth, subscription_id: str, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:

    url = f'https://{full_auth.domo_instance}.domo.com/api/publish/v2/subscription/{subscription_id}'

    res = await get_data(auth=full_auth,
                         method='POST',
                         url=url,
                         session=session,
                         debug=debug)
    return res


# Refreshing list of publish jobs. Typically "instance" = publisher instance


async def refresh_publish_jobs(full_auth: dmda.DomoFullAuth, publish_ids: list, session: aiohttp.ClientSession = None, debug: bool = False) -> rgd.ResponseGetData:

    url = f'https://{full_auth.domo_instance}.domo.com/api/publish/v2/publication/refresh'

    body = {
        'publicationIds': publish_ids
    }

    res = await get_data(auth=dev_auth,
                         method='PUT',
                         url=url,
                         body=body,
                         session=session,
                         debug=debug)
    return res



import asyncio
import aiohttp
from dataclasses import dataclass, field

from .routes import page_routes
from .DomoAuth import DomoDeveloperAuth, DomoFullAuth
from ..utils.DictDot import DictDot
from ..utils.Base import Base
from ..utils import convert
from typing import Any, List
from pprint import pprint
from datetime import datetime



@dataclass
class PageLayoutTemplate:
    content_key: int
    x: int
    y: int
    width: int
    height: int
    type: str
    virtual: bool
    virtual_appendix: bool
    
    @classmethod
    def _from_json(cls, dd):
        return cls(
            content_key=dd.contentKey,
            x=dd.x,
            y=dd.y,
            width=dd.width,
            height=dd.height,
            type= dd.type,
            virtual=dd.virtual,
            virtual_appendix = dd.virtualAppendix,
           
        )
    
    def get_body(self):
        return {
            "contentKey": self.content_key,
            "x": self.x,
            "y":self.y,
            "width": self.width,
            "height": self.height,
            "type":self.type,
            "virtual":self.virtual,
            "virtualAppendix":self.virtual_appendix
        }


@dataclass
class PageLayoutContent:
    accept_date_filter: bool
    accept_filters: bool
    accept_segments: bool
    card_id: int
    card_urn: str
    compact_interaction_default: bool
    content_key: int
    fit_to_frame: bool
    has_summary: bool
    hide_border: bool
    hide_description: bool
    hide_footer: bool
    hide_margins:bool
    hide_summary: bool
    hide_timeframe: bool
    hide_title: bool
    hide_wrench: bool
    id: int
    summary_number_only:bool
    type: str
    text: str
    
    
    
    @classmethod
    def _from_json(cls, dd):
        return cls(
            accept_date_filter=dd.acceptDateFilter,
            accept_filters=dd.acceptFilters,
            accept_segments=dd.acceptSegments,
            card_id=dd.cardId,
            card_urn=dd.cardUrn,
            compact_interaction_default= dd.compactInteractionDefault,
            content_key=dd.contentKey,
            fit_to_frame = dd.fitToFrame,
            has_summary=dd.hasSummary,
            hide_border = dd.hideBorder,
            hide_description= dd.hideDescription,
            hide_footer = dd.hideFooter,
            hide_margins = dd.hideMargins,
            hide_summary = dd.hideSummary,
            hide_timeframe = dd.hideTimeframe,
            hide_title = dd.hiteTitle,
            hide_wrench = dd.hideWrench,
            id = dd.id,
            summary_number_only=dd.summaryNumberOnly,
            type = dd.type,
            text = dd.text
        )
    
    def get_body(self):
        return {
            "acceptDateFilter": self.accept_date_filter,
            "acceptFilters": self.accept_filters,
            "acceptSegments":self.accept_segments,
            "cardId": self.card_id,
            "cardUrn": self.card_urn,
            "compactInteractionDefault":self.compact_interaction_default,
            "contentKey":self.content_key,
            "fitToFrame":self.fit_to_frame,
            "hasSummary":self.has_summary,
            "hideBorder":self.hide_border,
            "hideDescription":self.hide_description,
            "hideFooter":self.hide_footer,
            "hideMargins":self.hide_margins,
            "hideSummary":self.hide_summary,
            "hideTimeframe":self.hide_timeframe,
            "hiteTitle":self.hide_title,
            "hideWrench":self.hide_wrench,
            "id":self.id,
            "summaryNumberOnly":self.summary_number_only,
            "type":self.type,
            "text":self.text
        }
        
    
@dataclass
class PageLayoutStandard:
    aspect_ratio: float
    width: int
    frame_margin: int
    frame_padding: int
    type: str
    template: list[PageLayoutTemplate]
    
    @classmethod
    def _from_json(cls, dd):
        
        obj = cls(
            aspect_ratio=dd.aspectRatio,
            width=dd.width,
            frame_margin=dd.frameMargin,
            frame_padding=dd.framePadding,
            type = dd.type, 
            template = []
        )
            
        if dd.template is not None:
            for template_item in dd.template:
                dc = PageLayoutTemplate._from_json(
                    dd=template_item)
                if dc not in obj.template:
                    obj.template.append(dc)
        return obj
    
@dataclass
class PageLayoutCompact:
    aspect_ratio: float
    width: int
    frame_margin: int
    frame_padding: int
    type: str
    template: list[PageLayoutTemplate]
    
    @classmethod
    def _from_json(cls, dd):
        obj = cls(
            aspect_ratio=dd.aspectRatio,
            width=dd.width,
            frame_margin=dd.frameMargin,
            frame_padding=dd.framePadding,
            type = dd.type,
            template = []
        )
        if dd.template is not None:
            for template_item in dd.template:
                dc = PageLayoutTemplate._from_json(
                    dd=template_item)
                if dc not in obj.template:
                    obj.template.append(dc)
        return obj
    
@dataclass
class PageLayoutBackground:
    id: int
    crop_height: int
    crop_width: int
    x: int
    y: str
    data_file_id: int
    image_brightness: int
    image_height: int
    image_width: int
    selected_color: str
    text_color: str
    type: str
    is_darkMode: bool
    alpha: float
    src: str
    
    @classmethod
    def _from_json(cls, dd):
        if dd is not None:
            return cls(
            id=dd.id,
            crop_height=dd.cropHeight,
            crop_width=dd.cropWidth,
            x = dd.x,
            y = dd.y,
            data_file_id = dd.dataFileId,
            image_brightness = dd.imageBrightness,
            image_height = dd.imageHeight,
            image_width = dd.imageWidth,
            selected_color = dd.selectedColor,
            text_color = dd.textColor,
            type = dd.type,
            is_darkMode = dd.darkMode,
            alpha = dd.alpha,
            src = dd.src
            
            )
        else:
            return None


@dataclass
class PageLayout:
    id: str
    page_id: str
    is_print_friendly: bool
    is_enabled: bool
    is_dynamic: bool
    has_page_breaks : bool
    content : list[PageLayoutContent]
    standard : PageLayoutStandard
    compact : PageLayoutCompact
    background: PageLayoutBackground
    
    @classmethod
    def _from_json(cls, dd):
    
        obj= cls(
            id=dd.layoutId,
            page_id=dd.pageUrn,
            is_print_friendly=dd.printFriendly,
            is_enabled=dd.enabled,
            is_dynamic=dd.isDynamic,
            content = [],
            has_page_breaks = dd.hasPageBreaks,
            standard = PageLayoutStandard._from_json(
                    dd=dd.standard),
            compact = PageLayoutCompact._from_json(
                    dd=dd.compact),
            background =PageLayoutBackground._from_json(
                    dd=dd.background)
        )
        if dd.content is not None:
            for content_item in dd.content:
                dc = PageLayoutContent._from_json(
                    dd=content_item)
                if dc not in obj.content:
                    obj.content.append(dc)
        return obj
        
    
    @classmethod
    def generate_new_background_body (cls):
        background_body= {
         
            "selectedColor": "#EEE000",
            "textColor": "#4A4A4A",
            "type": "COLOR",
            "darkMode": False,
            "alpha": 1
        }
        
        return background_body

    
    def get_body (self):
        body = {
        "layoutId": self.id,
        "pageUrn": self.page_id,
        "printFriendly" : self.is_print_friendly,
        "enabled": self.is_enabled,
        "isDynamic": self.is_dynamic,
        "hasPageBreaks": self.has_page_breaks,
        "standard" : {
            "aspectRatio":self.standard.aspect_ratio,
            "width": self.standard.width,
            "frameMargin": self.standard.frame_margin,
            "framePadding": self.standard.frame_padding,
            "type": self.standard.type
            },
        "compact" : {
            "aspectRatio":self.compact.aspect_ratio,
            "width": self.compact.width,
            "frameMargin": self.compact.frame_margin,
            "framePadding": self.compact.frame_padding,
            "type": self.compact.type
            }
         }
        if self.background is not None:
            body["background"] = {
            "id": self.background.id,
            "cropHeight": self.background.crop_height,
            "cropWidth": self.background.crop_width,
            "x": self.background.x,
            "y": self.background.y,
            "dataFileId": self.background.data_file_id,
            "imageBrightness": self.background.image_brightness,
            "imageHeight": self.background.image_height,
            "imageWidth": self.background.image_width,
            "selectedColor": self.background.selected_color,
            "textColor": self.background.text_color,
            "type": self.background.type,
            "darkMode": self.background.is_darkMode,
            "alpha": self.background.alpha,
            "src": self.background.src
            }
        if  self.content ==[] or self.content is None:
            body["content"] = []
        else:
            temp_list = []
            for content_item in self.content:
                temp_list.append(content_item.get_body())
            body["content"]= temp_list
            
            
        if self.standard.template is None or self.standard.template==[]:
            body["standard"]["template"]=[]
        else:
            temp_list = []
            for template_item in self.standard.template:
                temp_list.append(template_item.get_body())
            body["standard"]["template"]= temp_list
            
        if self.compact.template is None or self.compact.template==[]:
            body["compact"]["template"]=[]
        else:
            temp_list = []
            for template_item in self.compact.template:
                temp_list.append(template_item.get_body())
            body["compact"]["template"]= temp_list
        return body

@dataclass
class DomoPage(Base):
    id: str
    title: str = None
    parent_page_id: str = None
    domo_instance: str = None
    full_auth: DomoFullAuth = None
    owners: list = field(default_factory=list)
    cards: list = field(default_factory=list)
    collections: list = field(default_factory=list)
    layout: PageLayout = None
    
    def __post_init__(self, full_auth=None):
        Base().__init__()

        if full_auth:
            self.domo_instance = full_auth.domo_instance
            self.full_auth = full_auth
        # self.Definition = CardDefinition(self)

    def display_url(self):
        return f'https://{self.domo_instance}.domo.com/page/{self.id}'

    @classmethod
    async def get_from_id(cls, page_id: str, full_auth: DomoFullAuth, debug: bool = False, include_layout: bool = False):
        import Library.DomoClasses.DomoCard as dc
        res = await page_routes.get_page_by_id(full_auth=full_auth, page_id=page_id, debug=debug, include_layout=include_layout)

        if res.status == 200:
            dd = DictDot(res.response)
            pg = cls(
                id=dd.id,
                domo_instance=full_auth.domo_instance,
                title=dd.title,
                parent_page_id=dd.page.parentPageId,
                owners=dd.page.owners,
                collections=dd.collections
            )
            if hasattr(dd, 'pageLayoutV4'):
                dd_layout = dd.pageLayoutV4
                pg.layout =PageLayout._from_json(
                    dd=dd.pageLayoutV4)

            pg.cards = await asyncio.gather(
                *[dc.DomoCard.get_from_id(id=card.id, full_auth=full_auth) for card in dd.cards])

            return pg

    @classmethod
    async def get_cards(cls, full_auth, page_id, debug: bool = False, session: aiohttp.ClientSession = None):
        try:
            import Library.DomoClasses.DomoCard as dc
            close_session = False if session else True

            if not session:
                session = aiohttp.ClientSession()

            res = await page_routes.get_page_definition(full_auth=full_auth, page_id=page_id, debug=debug, session=session)

            if res.status == 200:
                json = res.response

                card_list = [dc.DomoCard(id=card.get(
                    'id'), full_auth=full_auth) for card in json.get('cards')]

                return card_list

            else:
                return None

        finally:
            if close_session:
                await session.close()

    async def get_datasets(full_auth, page_id, debug: bool = False, session: aiohttp.ClientSession = None):
        try:
            import Library.DomoClasses.DomoDataset as dmds
            close_session = False if session else True

            if not session:
                session = aiohttp.ClientSession()

            res = await page_routes.get_page_definition(full_auth=full_auth, page_id=page_id, debug=debug, session=session)

            if res.status == 200:
                json = res.response

                dataset_ls = [card.get('datasources')
                              for card in json.get('cards')]

                return [dmds.DomoDataset(id=ds.get('dataSourceId'), full_auth=full_auth) for ds_ls in dataset_ls for ds in ds_ls]

            else:
                return None

        finally:
            if close_session:
                await session.close()

    @classmethod
    async def update_layout(cls,
                         full_auth: DomoFullAuth,
                         body: dict,
                         layout_id: str,
                         debug: bool = False,
                         log_results: bool = False,
                        session: aiohttp.ClientSession = None):
        datetime_now = datetime.now()
        start_time_epoch = convert.convert_datetime_to_epoch_millisecond(
            datetime_now)
        
        res_writelock = await page_routes.put_writelock(full_auth=full_auth, 
                                                        layout_id = layout_id,
                                                        user_id=full_auth.user_id,
                                                         epoch_time = start_time_epoch)
        if res_writelock.status == 200:
        
            res = await page_routes.update_page_layout(full_auth=full_auth,
                                          body=body,
                                       layout_id = layout_id,
                                       debug=debug,
                                       log_results=log_results)

    
            if res.status != 200:
                return False
            
            res_writelock = await page_routes.delete_writelock(full_auth=full_auth, 
                                                        layout_id = layout_id)
            if res_writelock.status != 200:
                return False
            
        else:
            return False

        return True
    


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/upload_data.ipynb.

# %% auto 0
__all__ = ['upload_data', 'upload_data_with_date']

# %% ../../nbs/utils/upload_data.ipynb 2
import httpx
import pandas as pd
import asyncio

import domolibrary.client.Logger as lc
import domolibrary.client.DomoAuth as dmda
import domolibrary.classes.DomoDataset as dmds


# %% ../../nbs/utils/upload_data.ipynb 3
async def loop_upload(
    upload_df: pd.DataFrame,
    consol_ds: dmds.DomoDataset,
    partition_key: str,
    upload_method: str,
    logger: lc.Logger,
    debug_api: bool = False,
    debug_prn: bool = False,
    debug_fn: bool = True,
    max_retry: int = 2,
    is_index: bool = False
):
    base_msg = f"{partition_key} in {consol_ds.auth.domo_instance}" if partition_key else f"in {consol_ds.auth.domo_instance}"

    if debug_fn:
        print(
            f"starting upload of {len(upload_df)} rows to {base_msg} with {max_retry} attempts")

    retry_attempt = 0

    while retry_attempt <= max_retry:
        try:
            retry_attempt += 1

            if debug_fn:
                print(f"attempt {retry_attempt} for {base_msg}")

            return await consol_ds.upload_data(
                upload_df=upload_df,
                upload_method="REPLACE" if partition_key else upload_method,
                partition_key=partition_key,
                is_index=is_index,
                debug_api=debug_api,
                debug_prn=debug_prn,
            )

        except Exception as e:
            print(e)

            if retry_attempt <= max_retry:
                message = f"⚠️ upload_data : unexpected error: {e} in {partition_key} during retry_attempt {retry_attempt}"

                logger.log_warning(message)
            else:
                message = f"💀⚠️ upload_data : failed to upload {partition_key} to {consol_ds.auth.domo_instance}"
                logger.log_error(message)
            
            print(message)


# %% ../../nbs/utils/upload_data.ipynb 4
async def upload_data(
    # instance where the data_fn function will execute against
    data_fn,  # data function to execute
    instance_auth: dmda.DomoAuth,  # instance to run the data function against
    consol_ds: dmds.DomoDataset,  # dataset where data should be accumulated
    # if partition key supplied, will replace existing partition
    partition_key: str = None,
    upload_method: str = 'REPLACE',
    is_index: bool = False,  # index dataset
    debug_prn: bool = False,
    debug_fn: bool = True,
    debug_api: bool = False,
    logger: lc.Logger = None,
    max_retry: int = 2  # number of times to attempt upload
):
    logger = logger or lc.Logger(app_name="upload_data")

    try:
        message = f"🏁 starting {instance_auth.domo_instance} - {data_fn.__name__}"

        if debug_fn:
            print(message)
        logger.log_info(message)

        instance_session = httpx.AsyncClient()

        upload_df = await data_fn(instance_auth, instance_session, debug_api=debug_api)

        if upload_df is None or len(upload_df.index) == 0:
            print("no data to upload")
            return None

        if debug_fn:
            print(upload_df.head())

        res = await loop_upload(
            upload_df=upload_df,
            consol_ds=consol_ds,
            partition_key=partition_key,
            upload_method=upload_method,
            debug_api=debug_api,
            debug_prn=debug_prn,
            debug_fn=debug_fn,
            max_retry=max_retry,
            logger=logger,
            is_index=False
        )

        if res.is_success:
                message = f"🚀 success upload of {partition_key} to {consol_ds.id} in {consol_ds.auth.domo_instance} in {data_fn.__name__}"

        else:
            message = f"💣 upload_data successful status but failed to upload {partition_key} - {res.status} - {res.response} in {data_fn.__name__}"
            
        if debug_fn:
            print(message)
        
        logger.log_info(message)
        
        return res

    finally:
        if is_index:

            res = await consol_ds.index_dataset(debug_api=debug_api, session=instance_session)
            if res.is_success:
                message = f"🥫 successfully indexed {consol_ds.name} in {consol_ds.auth.domo_instance}"
            else:
                message = f"💀⚠️ failure to index {consol_ds.name} in {consol_ds.auth.domo_instance}"
            if debug_fn:
                print(message)
            logger.log_info(message)
        await instance_session.aclose()


# %% ../../nbs/utils/upload_data.ipynb 7
async def upload_data_with_date(
    instance_auth,
    consol_auth,
    data_fn,
    consol_ds,
    partition_date_col,
    partition_delimiter,
    start_date,
    end_date,
    debug: bool = False,
    debug_prn: bool = False,
):

    instance_session = httpx.AsyncClient()

    print(
        f"'🎬 upload_with_data: starting retrieval {start_date}, {end_date}, {instance_auth.domo_instance}"
    )

    upload_df = await data_fn(
        instance_auth=instance_auth,
        session=instance_session,
        start_date=start_date,
        end_date=end_date,
        debug=debug,
    )

    await instance_session.aclose()

    if not isinstance(upload_df, pd.DataFrame):
        print(f"🛑 error no data returned {instance_auth.domo_instance}")
        print(upload_df)
        return None

    if debug_prn:
        print(
            f"🧻 upload_with_data: starting upload {len(upload_df)} rows for {instance_auth.domo_instance}"
        )

    task = []

    for index, partition_set in upload_df.drop_duplicates(
        subset=[partition_date_col]
    ).iterrows():
        partition_date = partition_set[partition_date_col]

        partition_key = (
            f"{instance_auth.domo_instance}{partition_delimiter}{str(partition_date)}"
        )

        task.append(
            consol_ds.upload_data(
                upload_df=upload_df[(upload_df[partition_date_col] == partition_date)],
                upload_method="REPLACE",
                partition_key=partition_key,
                is_index=False,
                debug_api=debug_api,
                debug_prn=debug_prn,
            )
        )

    res = await asyncio.gather(*task)

    if debug_prn:
        print(
            f"🎉 upload_with_data : finished uploading {len(upload_df)} rows for {instance_auth.domo_instance}"
        )
    return res
